@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997-1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -37,16 +37,35 @@
  */
 
 #include "rsh_locl.h"
-RCSID("$Id: rshd.c,v 1.12 1997/10/29 01:25:14 assar Exp $");
+RCSID("$Id: rshd.c,v 1.20.10.1 1999/07/26 14:19:26 assar Exp $");
 
 enum auth_method auth_method;
 
 krb5_context context;
 krb5_keyblock *keyblock;
+krb5_crypto crypto;
 des_key_schedule schedule;
 des_cblock iv;
 
-int do_encrypt;
+int do_encrypt = 0;
+
+static int do_unique_tkfile           = 0;
+static char unique_tkfile[MAXPATHLEN] = "";
+static krb5_data saved_remote_cred;
+static krb5_auth_context saved_auth_context;
+static krb5_principal saved_client;
+static char saved_ccname[MAXPATHLEN];
+static int save_status                = 1;
+
+static int do_inetd = 1;
+static char *port_str;
+static int do_rhosts;
+static int do_kerberos = 0;
+static int do_vacuous = 0;
+static int do_log = 1;
+static int do_newpag = 1;
+static int do_version;
+static int do_help = 0;
 
 static void
 syslog_and_die (const char *m, ...)
@@ -89,6 +108,28 @@ read_str (int s, char *str, size_t sz, char *expl)
     fatal (s, "%s too long", expl);
 }
 
+static int
+recv_bsd_auth (int s, u_char *buf,
+	       struct sockaddr_in thisaddr,
+	       struct sockaddr_in thataddr,
+	       char *client_username,
+	       char *server_username,
+	       char *cmd)
+{
+    struct passwd *pwd;
+
+    read_str (s, client_username, USERNAME_SZ, "local username");
+    read_str (s, server_username, USERNAME_SZ, "remote username");
+    read_str (s, cmd, COMMAND_SZ, "command");
+    pwd = getpwnam(server_username);
+    if (pwd == NULL)
+	fatal(s, "Login incorrect.");
+    if (iruserok(thataddr.sin_addr.s_addr, pwd->pw_uid == 0,
+		 client_username, server_username))
+	fatal(s, "Login incorrect.");
+    return 0;
+}
+
 #ifdef KRB4
 static int
 recv_krb4_auth (int s, u_char *buf,
@@ -150,45 +191,50 @@ recv_krb4_auth (int s, u_char *buf,
 
 #endif /* KRB4 */
 
+static int 
+save_krb5_creds (int s,
+                 krb5_auth_context auth_context,
+                 krb5_principal client)
+
+{
+    saved_auth_context = auth_context;
+    saved_client = client;
+    krb5_data_zero (&saved_remote_cred);
+    return krb5_read_message (context,
+			      (void *)&s,
+			      &saved_remote_cred);
+}
+
 static void
-recv_krb5_creds (int s,
-		 krb5_auth_context auth_context,
-		 char *username,
-		 krb5_principal client)
+recv_krb5_creds (krb5_auth_context auth_context,
+		 krb5_principal *client,
+                 krb5_data *remote_cred,
+                 char *ccname)
 {
     krb5_error_code ret;
     krb5_ccache ccache;
-    krb5_data data;
-    char ccname[MAXPATHLEN];
-    struct passwd *pwd;
 
-    krb5_data_zero (&data);
-    ret = krb5_read_message (context,
-			     (void *)&s,
-			     &data);
-    if (ret || data.length == 0)
+    if (remote_cred->length == 0)
 	return;
 
-    pwd = getpwnam (username);
-    if (pwd == NULL)
-	goto out;
-
-    snprintf (ccname, sizeof(ccname),
-	      "FILE:/tmp/krb5cc_%u", pwd->pw_uid);
     ret = krb5_cc_resolve (context, ccname, &ccache);
-    if (ret)
+    if (ret) {
+	krb5_warn (context, ret, "krb5_cc_resolve");
 	goto out;
-    ret = krb5_cc_initialize (context, ccache, client);
-    if (ret)
+    }
+    ret = krb5_cc_initialize (context, ccache, *client);
+    if (ret) {
+	krb5_warn (context, ret, "krb5_cc_initialize");
 	goto out;
-    ret = krb5_rd_cred (context, auth_context, ccache, &data);
+    }
+    ret = krb5_rd_cred(context, auth_context, ccache, remote_cred);
+    if (ret) {
+	krb5_warn (context, ret, "krb5_rd_cred");
+    }
     krb5_cc_close (context, ccache);
-    if (ret)
-	goto out;
-    chown (ccname + 5, pwd->pw_uid, -1);
 
 out:
-    krb5_data_free (&data);
+    krb5_data_free (remote_cred);
 }
 
 static int
@@ -203,7 +249,6 @@ recv_krb5_auth (int s, u_char *buf,
     krb5_auth_context auth_context = NULL;
     krb5_ticket *ticket;
     krb5_error_code status;
-    krb5_authenticator authenticator;
     krb5_data cksum_data;
     krb5_principal server;
 
@@ -241,41 +286,83 @@ recv_krb5_auth (int s, u_char *buf,
 	syslog_and_die ("krb5_recvauth: %s",
 			krb5_get_err_text(context, status));
 
-    read_str (s, client_username, USERNAME_SZ, "local username");
-    read_str (s, cmd, COMMAND_SZ, "command");
     read_str (s, server_username, USERNAME_SZ, "remote username");
+    read_str (s, cmd, COMMAND_SZ, "command");
+    read_str (s, client_username, COMMAND_SZ, "local username");
 
     status = krb5_auth_con_getkey (context, auth_context, &keyblock);
     if (status)
-	syslog_and_die ("krb5_auth_con_getkey: %s",
-			krb5_get_err_text(context, status));
+       syslog_and_die ("krb5_auth_con_getkey: %s",
+                       krb5_get_err_text(context, status));
 
-    status = krb5_auth_getauthenticator (context,
-					 auth_context,
-					 &authenticator);
-    if (status)
-	syslog_and_die ("krb5_auth_getauthenticator: %s",
-			krb5_get_err_text(context, status));
+    status = krb5_crypto_init(context, keyblock, 0, &crypto);
+    if(status)
+	syslog_and_die("krb5_crypto_init: %s", 
+		       krb5_get_err_text(context, status));
 
+    
     cksum_data.length = asprintf ((char **)&cksum_data.data,
 				  "%u:%s%s",
 				  ntohs(thisaddr.sin_port),
 				  cmd,
 				  server_username);
 
-    status = krb5_verify_checksum (context,
-				   cksum_data.data,
-				   cksum_data.length,
-				   keyblock,
-				   authenticator->cksum);
+    status = krb5_verify_authenticator_checksum(context, 
+						auth_context,
+						cksum_data.data, 
+						cksum_data.length);
+
     if (status)
-	syslog_and_die ("krb5_verify_checksum: %s",
+	syslog_and_die ("krb5_verify_authenticator_checksum: %s",
 			krb5_get_err_text(context, status));
 
     free (cksum_data.data);
-    krb5_free_authenticator (context, &authenticator);
 
-    recv_krb5_creds (s, auth_context, server_username, ticket->client);
+    if (strncmp (client_username, "-u ", 3) == 0) {
+	do_unique_tkfile = 1;
+	memmove (client_username, client_username + 3,
+		 strlen(client_username) - 2);
+    }
+
+    if (strncmp (server_username, "-U ", 3) == 0) {
+	char *end;
+
+	do_unique_tkfile = 1;
+	end = strchr(server_username + 3,' ');
+	strncpy(unique_tkfile, server_username + 3, end - server_username - 3);
+	unique_tkfile[end - server_username - 3] = '\0';
+	memmove (server_username, end +1, strlen(end+1)+1);
+    }
+
+    {
+	struct passwd *test_pwd; 
+	test_pwd = getpwnam (server_username);
+	if (test_pwd == NULL)
+	    fatal (s, "Login incorrect.");
+
+	if (do_unique_tkfile) {
+	    if (strcmp(unique_tkfile,"") == 0) {
+		int f;
+		char tf[MAXPATHLEN];
+          
+		do { 
+		    snprintf(tf, sizeof(tf), "%s%u_%u", KRB5_DEFAULT_CCROOT,
+			     test_pwd->pw_uid,
+			     (unsigned int)(getpid()*time(0)));
+		    f = open(tf + strlen("FILE:"), O_CREAT|O_EXCL|O_RDWR);
+		} while(f < 0); 
+		close(f);
+		unlink(tf + strlen("FILE:"));
+		strcpy(saved_ccname,tf);
+           } else
+	       snprintf (saved_ccname, sizeof(saved_ccname),
+			 "FILE:%s",unique_tkfile);
+	} else
+	    snprintf (saved_ccname, sizeof(saved_ccname),
+		      "FILE:/tmp/krb5cc_%u", test_pwd->pw_uid);
+    } 
+
+    save_status = save_krb5_creds (s, auth_context, ticket->client);
 
     if(!krb5_kuserok (context,
 		     ticket->client,
@@ -285,6 +372,8 @@ recv_krb5_auth (int s, u_char *buf,
     if (strncmp (cmd, "-x ", 3) == 0) {
 	do_encrypt = 1;
 	memmove (cmd, cmd + 3, strlen(cmd) - 2);
+    } else {
+	do_encrypt = 0;
     }
 
     {
@@ -309,7 +398,7 @@ loop (int from0, int to0,
       int to1,   int from1,
       int to2,   int from2)
 {
-    struct fd_set real_readset;
+    fd_set real_readset;
     int max_fd;
     int count = 2;
 
@@ -320,15 +409,16 @@ loop (int from0, int to0,
     max_fd = max(from0, max(from1, from2)) + 1;
     for (;;) {
 	int ret;
-	struct fd_set readset = real_readset;
+	fd_set readset = real_readset;
 	char buf[RSH_BUFSIZ];
 
 	ret = select (max_fd, &readset, NULL, NULL, NULL);
-	if (ret < 0)
+	if (ret < 0) {
 	    if (errno == EINTR)
 		continue;
 	    else
 		syslog_and_die ("select: %m");
+	}
 	if (FD_ISSET(from0, &readset)) {
 	    ret = do_read (from0, buf, sizeof(buf));
 	    if (ret < 0)
@@ -366,18 +456,35 @@ loop (int from0, int to0,
    }
 }
 
+/*
+ * Used by `setup_copier' to create some pipe-like means of
+ * communcation.  Real pipes would probably be the best thing, but
+ * then the shell doesn't understand it's talking to rshd.  If
+ * socketpair doesn't work everywhere, some autoconf magic would have
+ * to be added here.
+ *
+ * If it fails creating the `pipe', it aborts by calling fatal.
+ */
+
+static void
+pipe_a_like (int fd[2])
+{
+    if (socketpair (AF_UNIX, SOCK_STREAM, 0, fd) < 0)
+	fatal (STDOUT_FILENO, "socketpair: %m");
+}
+
+/*
+ * Start a child process and leave the parent copying data to and from it.  */
+
 static void
 setup_copier (void)
 {
     int p0[2], p1[2], p2[2];
     pid_t pid;
 
-    if (pipe(p0) < 0)
-	fatal (STDOUT_FILENO, "pipe: %m");
-    if (pipe(p1) < 0)
-	fatal (STDOUT_FILENO, "pipe: %m");
-    if (pipe(p2) < 0)
-	fatal (STDOUT_FILENO, "pipe: %m");
+    pipe_a_like(p0);
+    pipe_a_like(p1);
+    pipe_a_like(p2);
     pid = fork ();
     if (pid < 0)
 	fatal (STDOUT_FILENO, "fork: %m");
@@ -391,7 +498,7 @@ setup_copier (void)
 	close (p0[0]);
 	close (p1[1]);
 	close (p2[1]);
-    } else {
+    } else { /* parent */
 	close (p0[0]);
 	close (p1[1]);
 	close (p2[1]);
@@ -405,8 +512,36 @@ setup_copier (void)
     }
 }
 
+/*
+ * Is `port' a ``reserverd'' port?
+ */
+
+static int
+is_reserved(u_short port)
+{
+    return ntohs(port) < IPPORT_RESERVED;
+}
+
+/*
+ * Set the necessary part of the environment in `env'.
+ */
+
 static void
-doit (void)
+setup_environment (char *env[7], struct passwd *pwd)
+{
+    asprintf (&env[0], "USER=%s",  pwd->pw_name);
+    asprintf (&env[1], "HOME=%s",  pwd->pw_dir);
+    asprintf (&env[2], "SHELL=%s", pwd->pw_shell);
+    asprintf (&env[3], "PATH=%s",  _PATH_DEFPATH);
+    asprintf (&env[4], "SSH_CLIENT=only_to_make_bash_happy");
+    if (do_unique_tkfile)
+       asprintf (&env[5], "KRB5CCNAME=%s",saved_ccname);
+       else env[5] = NULL;
+    env[6] = NULL;
+}
+
+static void
+doit (int do_kerberos, int check_rhosts)
 {
     u_char buf[BUFSIZ];
     u_char *p;
@@ -414,11 +549,11 @@ doit (void)
     int addrlen;
     int port;
     int errsock = -1;
-    char client_user[16], server_user[16];
+    char client_user[COMMAND_SZ], server_user[USERNAME_SZ];
     char cmd[COMMAND_SZ];
     struct passwd *pwd;
     int s = STDIN_FILENO;
-    char *env[5];
+    char *env[7];
 
     addrlen = sizeof(thisaddr);
     if (getsockname (s, (struct sockaddr *)&thisaddr, &addrlen) < 0
@@ -431,6 +566,9 @@ doit (void)
 	syslog_and_die ("getpeername: %m");
     }
 
+    if (!do_kerberos && !is_reserved(thataddr.sin_port))
+	fatal(s, "Permission denied");
+
     p = buf;
     port = 0;
     for(;;) {
@@ -443,6 +581,10 @@ doit (void)
 	else
 	    syslog_and_die ("non-digit in port number: %c", *p);
     }
+
+    if (!do_kerberos && !is_reserved(htons(port)))
+	fatal(s, "Permission denied");
+
     if (port) {
 	int priv_port = IPPORT_RESERVED - 1;
 
@@ -463,25 +605,39 @@ doit (void)
 	    syslog_and_die ("connect: %m");
     }
     
-    if (net_read (s, buf, 4) != 4)
-	syslog_and_die ("reading auth info: %m");
+    if(do_kerberos) {
+	if (net_read (s, buf, 4) != 4)
+	    syslog_and_die ("reading auth info: %m");
     
 #ifdef KRB4
-    if (recv_krb4_auth (s, buf, thisaddr, thataddr,
-			client_user,
-			server_user,
-			cmd) == 0)
-	auth_method = AUTH_KRB4;
-    else
-#endif /* KRB4 */
-    if(recv_krb5_auth (s, buf, thisaddr, thataddr,
+	if (recv_krb4_auth (s, buf, thisaddr, thataddr,
 			    client_user,
 			    server_user,
 			    cmd) == 0)
-	auth_method = AUTH_KRB5;
-    else
-	syslog_and_die ("unrecognized auth protocol: %x %x %x %x",
-			buf[0], buf[1], buf[2], buf[3]);
+	    auth_method = AUTH_KRB4;
+	else
+#endif /* KRB4 */
+	    if(recv_krb5_auth (s, buf, thisaddr, thataddr,
+			       client_user,
+			       server_user,
+			       cmd) == 0)
+		auth_method = AUTH_KRB5;
+	    else
+		syslog_and_die ("unrecognized auth protocol: %x %x %x %x",
+				buf[0], buf[1], buf[2], buf[3]);
+    } else {
+	if(recv_bsd_auth (s, buf, thisaddr, thataddr,
+			  client_user,
+			  server_user,
+			  cmd) == 0) {
+	    auth_method = AUTH_BROKEN;
+	    if(do_vacuous) {
+		printf("Remote host requires Kerberos authentication\n");
+		exit(0);
+	    }
+	} else
+	    syslog_and_die("recv_bsd_auth failed");
+    }
 
     pwd = getpwnam (server_user);
     if (pwd == NULL)
@@ -492,6 +648,19 @@ doit (void)
 
     if (pwd->pw_uid != 0 && access (_PATH_NOLOGIN, F_OK) == 0)
 	fatal (s, "Login disabled.");
+
+#ifdef HAVE_GETSPNAM
+    {
+	struct spwd *sp;
+	long    today;
+    
+	sp = getspnam(server_user);
+	today = time(0)/(24L * 60 * 60);
+	if (sp->sp_expire > 0) 
+	    if (today > sp->sp_expire) 
+		fatal(s, "Account has expired.");
+    }
+#endif
     
 #ifdef HAVE_SETLOGIN
     if (setlogin(pwd->pw_name) < 0)
@@ -505,9 +674,18 @@ doit (void)
     if (initgroups (pwd->pw_name, pwd->pw_gid) < 0)
 	fatal (s, "Login incorrect.");
 
+    if (setgid(pwd->pw_gid) < 0)
+	fatal (s, "Login incorrect.");
+
     if (setuid (pwd->pw_uid) < 0)
 	fatal (s, "Login incorrect.");
 
+#ifdef KRB5
+    if (!save_status)
+        recv_krb5_creds(saved_auth_context, &saved_client,
+                        &saved_remote_cred,saved_ccname);
+#endif
+
     if (chdir (pwd->pw_dir) < 0)
 	fatal (s, "Remote directory.");
 
@@ -517,11 +695,7 @@ doit (void)
 	close (errsock);
     }
 
-    asprintf (&env[0], "USER=%s",  pwd->pw_name);
-    asprintf (&env[1], "HOME=%s",  pwd->pw_dir);
-    asprintf (&env[2], "SHELL=%s", pwd->pw_shell);
-    asprintf (&env[3], "PATH=%s",  _PATH_DEFPATH);
-    env[4] = NULL;
+    setup_environment (env, pwd);
 
     if (do_encrypt) {
 	setup_copier ();
@@ -530,63 +704,137 @@ doit (void)
 	    fatal (s, "write failed");
     }
 
+#ifdef KRB4
+    if(k_hasafs()) {
+	char cell[64];
+	char *pw_dir;
+
+#ifdef _AIX
+	/* XXX this is a fix for a bug in AFS for AIX 4.3, w/o
+	   this hack the kernel crashes on the following
+	   pioctl... */
+	pw_dir = strdup(pwd->pw_dir);
+#else
+	pw_dir = pwd->pw_dir;
+#endif
+
+	if(do_newpag)
+	    k_setpag();
+	if (k_afs_cell_of_file (pw_dir, cell, sizeof(cell)) == 0)
+	    krb_afslog_uid_home (cell, NULL, pwd->pw_uid, pwd->pw_dir);
+
+	krb_afslog_uid_home(NULL, NULL, pwd->pw_uid, pwd->pw_dir);
+
+#ifdef KRB5
+	/* XXX */
+       {
+	   krb5_ccache ccache;
+	   krb5_error_code status;
+
+	   if (do_unique_tkfile)
+	       setenv("KRB5CCNAME", saved_ccname, 1);
+	   status = krb5_cc_default (context, &ccache);
+	   if (!status) {
+	       krb5_afslog_uid_home(context,ccache,NULL,NULL,
+				    pwd->pw_uid, pwd->pw_dir);
+	       krb5_cc_close (context, ccache);
+	   }
+       }
+#endif /* KRB5 */
+    }
+#endif /* KRB4 */
     execle (pwd->pw_shell, pwd->pw_shell, "-c", cmd, NULL, env);
     err(1, "exec %s", pwd->pw_shell);
 }
 
+struct getargs args[] = {
+    { "inetd",		'i',	arg_negative_flag,	&do_inetd,
+      "Not started from inetd" },
+    { "kerberos",	'k',	arg_flag,	&do_kerberos,
+      "Implement kerberised services" },
+    { "encrypt",	'x',	arg_flag,		&do_encrypt,
+      "Implement encrypted service" },
+    { "rhosts",		'l',	arg_flag, &do_rhosts,
+      "Check users .rhosts" },
+    { "port",		'p',	arg_string,	&port_str,	"Use this port",
+      "port" },
+    { "vacuous",	'v',	arg_flag, &do_vacuous,
+      "Don't accept non-kerberised connections" },
+    { NULL,		'P',	arg_negative_flag, &do_newpag,
+      "Don't put process in new PAG" },
+    /* compatibility flag: */
+    { NULL,		'L',	arg_flag, &do_log },
+    { "version",	0, 	arg_flag,		&do_version },
+    { "help",		0, 	arg_flag,		&do_help }
+};
+
 static void
-usage (void)
+usage (int ret)
 {
-    syslog (LOG_ERR, "Usage: %s [-ix] [-p port]", __progname);
-    exit (1);
+    if(isatty(STDIN_FILENO))
+	arg_printusage (args,
+			sizeof(args) / sizeof(args[0]),
+			NULL,
+			"");
+    else
+	syslog (LOG_ERR, "Usage: %s [-ikxlvPL] [-p port]", __progname);
+    exit (ret);
 }
 
+
 int
 main(int argc, char **argv)
 {
-    int c;
-    int inetd = 0;
+    int optind = 0;
     int port = 0;
 
     set_progname (argv[0]);
     roken_openlog ("rshd", LOG_ODELAY | LOG_PID, LOG_AUTH);
 
-    while ((c = getopt(argc, argv, "ixp:")) != EOF) {
-	switch (c) {
-	case 'i' :
-	    inetd = 1;
-	    break;
-	case 'x' :
-	    do_encrypt = 1;
-	    break;
-	case 'p': {
-	    struct servent *s = roken_getservbyname (optarg, "tcp");
+    if (getarg(args, sizeof(args) / sizeof(args[0]), argc, argv,
+	       &optind))
+	usage(1);
 
-	    if (s)
-		port = s->s_port;
-	    else {
-		char *ptr;
+    if(do_help)
+	usage (0);
 
-		port = strtol (optarg, &ptr, 10);
-		if (port == 0 && ptr == optarg)
-		    syslog_and_die ("Bad port `%s'", optarg);
-		port = htons(port);
-	    }
-	    break;
-	}
-	default :
-	    usage ();
+    if (do_version) {
+	print_version(NULL);
+	exit(0);
+    }
+
+    if(port_str) {
+	struct servent *s = roken_getservbyname (port_str, "tcp");
+
+	if (s)
+	    port = s->s_port;
+	else {
+	    char *ptr;
+
+	    port = strtol (port_str, &ptr, 10);
+	    if (port == 0 && ptr == port_str)
+		syslog_and_die("Bad port `%s'", port_str);
+	    port = htons(port);
 	}
     }
-    if (inetd) {
-	if (port == 0)
-	    if (do_encrypt)
-		port = krb5_getportbyname (context, "ekshell", "tcp", 545);
-	    else
-		port = krb5_getportbyname (context, "kshell",  "tcp", 544);
+
+    if (do_encrypt)
+	do_kerberos = 1;
+
+    if (!do_inetd) {
+	if (port == 0) {
+	    if (do_kerberos) {
+		if (do_encrypt)
+		    port = krb5_getportbyname (context, "ekshell", "tcp", 545);
+		else
+		    port = krb5_getportbyname (context, "kshell",  "tcp", 544);
+	    } else {
+		port = krb5_getportbyname(context, "shell", "tcp", 514);
+	    }
+	}
 	mini_inetd (port);
     }
 
-    doit ();
+    doit (do_kerberos, do_rhosts);
     return 0;
 }
