@@ -33,7 +33,7 @@
 
 #include "telnetd.h"
 
-RCSID("$Id: telnetd.c,v 1.47 1997/10/29 01:26:58 assar Exp $");
+RCSID("$Id: telnetd.c,v 1.54 1999/07/26 00:44:13 assar Exp $");
 
 #ifdef _SC_CRAY_SECURE_SYS
 #include <sys/sysv.h>
@@ -117,7 +117,7 @@ int debug = 0;
 int keepalive = 1;
 char *progname;
 
-extern void usage (void);
+static void usage (void);
 
 /*
  * The string to pass to getopt().  We do it this way so
@@ -136,12 +136,14 @@ char valid_opts[] = "Bd:hklnS:u:UL:y"
 #endif
 		    ;
 
-void doit(struct sockaddr_in*);
+static void doit(struct sockaddr*, int);
 
-int main(int argc, char **argv)
+int
+main(int argc, char **argv)
 {
-    struct sockaddr_in from;
-    int on = 1, fromlen;
+    struct sockaddr_storage __ss;
+    struct sockaddr *sa = (struct sockaddr *)&__ss;
+    int on = 1, sa_size;
     int ch;
 #if	defined(IPPROTO_IP) && defined(IP_TOS)
     int tos = -1;
@@ -275,7 +277,7 @@ int main(int argc, char **argv)
 #endif	/* CRAY */
 
 	case 'S':
-#ifdef	HAS_GETTOS
+#ifdef	HAVE_PARSETOS
 	    if ((tos = parsetos(optarg, "tcp")) < 0)
 		fprintf(stderr, "%s%s%s\n",
 			"telnetd: Bad TOS argument '", optarg,
@@ -329,7 +331,7 @@ int main(int argc, char **argv)
     argv += optind;
 
     if (debug) {
-	int port;
+	int port = 0;
 	struct servent *sp;
 
 	if (argc > 1) {
@@ -366,10 +368,8 @@ int main(int argc, char **argv)
 
 	memset(&dv, 0, sizeof(dv));
 
-	if (getsysv(&sysv, sizeof(struct sysv)) != 0) {
-	    perror("getsysv");
-	    exit(1);
-	}
+	if (getsysv(&sysv, sizeof(struct sysv)) != 0) 
+	    fatalperror(net, "getsysv");
 
 	/*
 	 *	Get socket security label and set device values
@@ -379,10 +379,9 @@ int main(int argc, char **argv)
 	if ((getsockopt(0, SOL_SOCKET, SO_SECURITY,
 			(void *)&ss, &szss) < 0) ||
 	    (getsockopt(0, SOL_SOCKET, SO_SEC_MULTI,
-			(void *)&sock_multi, &szi) < 0)) {
-	    perror("getsockopt");
-	    exit(1);
-	} else {
+			(void *)&sock_multi, &szi) < 0)) 
+	    fatalperror(net, "getsockopt");
+	else {
 	    dv.dv_actlvl = ss.ss_actlabel.lt_level;
 	    dv.dv_actcmp = ss.ss_actlabel.lt_compart;
 	    if (!sock_multi) {
@@ -409,21 +408,21 @@ int main(int argc, char **argv)
 #endif	/* _SC_CRAY_SECURE_SYS */
 
     roken_openlog("telnetd", LOG_PID | LOG_ODELAY, LOG_DAEMON);
-    fromlen = sizeof (from);
-    if (getpeername(0, (struct sockaddr *)&from, &fromlen) < 0) {
+    sa_size = sizeof (__ss);
+    if (getpeername(STDIN_FILENO, sa, &sa_size) < 0) {
 	fprintf(stderr, "%s: ", progname);
 	perror("getpeername");
 	_exit(1);
     }
     if (keepalive &&
-	setsockopt(0, SOL_SOCKET, SO_KEEPALIVE,
+	setsockopt(STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE,
 		   (void *)&on, sizeof (on)) < 0) {
 	syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
     }
 
 #if	defined(IPPROTO_IP) && defined(IP_TOS) && defined(HAVE_SETSOCKOPT)
     {
-# ifdef HAS_GETTOS
+# ifdef HAVE_GETTOSBYNAME
 	struct tosent *tp;
 	if (tos < 0 && (tp = gettosbyname("telnet", "tcp")))
 	    tos = tp->t_tos;
@@ -431,19 +430,21 @@ int main(int argc, char **argv)
 	if (tos < 0)
 	    tos = 020;	/* Low Delay bit */
 	if (tos
-	    && (setsockopt(0, IPPROTO_IP, IP_TOS,
+	    && sa->sa_family == AF_INET
+	    && (setsockopt(STDIN_FILENO, IPPROTO_IP, IP_TOS,
 			   (void *)&tos, sizeof(tos)) < 0)
 	    && (errno != ENOPROTOOPT) )
 	    syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
     }
 #endif	/* defined(IPPROTO_IP) && defined(IP_TOS) */
-    net = 0;
-    doit(&from);
+    net = STDIN_FILENO;
+    doit(sa, sa_size);
     /* NOTREACHED */
+    return 0;
 }  /* end of main */
 
-void
-usage()
+static void
+usage(void)
 {
     fprintf(stderr, "Usage: telnetd");
 #ifdef	AUTHENTICATION
@@ -463,7 +464,7 @@ usage()
     fprintf(stderr, " [-r[lowpty]-[highpty]]");
 #endif
     fprintf(stderr, "\n\t");
-#ifdef	HAS_GETTOS
+#ifdef	HAVE_GETTOSBYNAME
     fprintf(stderr, " [-S tos]");
 #endif
 #ifdef	AUTHENTICATION
@@ -485,7 +486,7 @@ static unsigned char ttytype_sbbuf[] = {
 };
 
 int
-getterminaltype(char *name)
+getterminaltype(char *name, size_t name_sz)
 {
     int retval = -1;
     void _gettermname();
@@ -499,7 +500,7 @@ getterminaltype(char *name)
     while (his_will_wont_is_changing(TELOPT_AUTHENTICATION))
 	ttloop();
     if (his_state_is_will(TELOPT_AUTHENTICATION)) {
-	retval = auth_wait(name);
+	retval = auth_wait(name, name_sz);
     }
 #endif
 
@@ -536,33 +537,33 @@ getterminaltype(char *name)
 	static unsigned char sb[] =
 	{ IAC, SB, TELOPT_TSPEED, TELQUAL_SEND, IAC, SE };
 
-	net_write (sb, sizeof sb);
+	telnet_net_write (sb, sizeof sb);
 	DIAG(TD_OPTIONS, printsub('>', sb + 2, sizeof sb - 2););
     }
     if (his_state_is_will(TELOPT_XDISPLOC)) {
 	static unsigned char sb[] =
 	{ IAC, SB, TELOPT_XDISPLOC, TELQUAL_SEND, IAC, SE };
 
-	net_write (sb, sizeof sb);
+	telnet_net_write (sb, sizeof sb);
 	DIAG(TD_OPTIONS, printsub('>', sb + 2, sizeof sb - 2););
     }
     if (his_state_is_will(TELOPT_NEW_ENVIRON)) {
 	static unsigned char sb[] =
 	{ IAC, SB, TELOPT_NEW_ENVIRON, TELQUAL_SEND, IAC, SE };
 
-	net_write (sb, sizeof sb);
+	telnet_net_write (sb, sizeof sb);
 	DIAG(TD_OPTIONS, printsub('>', sb + 2, sizeof sb - 2););
     }
     else if (his_state_is_will(TELOPT_OLD_ENVIRON)) {
 	static unsigned char sb[] =
 	{ IAC, SB, TELOPT_OLD_ENVIRON, TELQUAL_SEND, IAC, SE };
 
-	net_write (sb, sizeof sb);
+	telnet_net_write (sb, sizeof sb);
 	DIAG(TD_OPTIONS, printsub('>', sb + 2, sizeof sb - 2););
     }
     if (his_state_is_will(TELOPT_TTYPE)) {
 
-	net_write (ttytype_sbbuf, sizeof ttytype_sbbuf);
+	telnet_net_write (ttytype_sbbuf, sizeof ttytype_sbbuf);
 	DIAG(TD_OPTIONS, printsub('>', ttytype_sbbuf + 2,
 				  sizeof ttytype_sbbuf - 2););
     }
@@ -593,12 +594,12 @@ getterminaltype(char *name)
 	 * we have to just go with what we (might) have already gotten.
 	 */
 	if (his_state_is_will(TELOPT_TTYPE) && !terminaltypeok(terminaltype)) {
-	    strncpy(first, terminaltype, sizeof(first));
+	    strcpy_truncate(first, terminaltype, sizeof(first));
 	    for(;;) {
 		/*
 		 * Save the unknown name, and request the next name.
 		 */
-		strncpy(last, terminaltype, sizeof(last));
+		strcpy_truncate(last, terminaltype, sizeof(last));
 		_gettermname();
 		if (terminaltypeok(terminaltype))
 		    break;
@@ -637,7 +638,7 @@ _gettermname()
     if (his_state_is_wont(TELOPT_TTYPE))
 	return;
     settimer(baseline);
-    net_write (ttytype_sbbuf, sizeof ttytype_sbbuf);
+    telnet_net_write (ttytype_sbbuf, sizeof ttytype_sbbuf);
     DIAG(TD_OPTIONS, printsub('>', ttytype_sbbuf + 2,
 			      sizeof ttytype_sbbuf - 2););
     while (sequenceIs(ttypesubopt, baseline))
@@ -647,21 +648,7 @@ _gettermname()
 int
 terminaltypeok(char *s)
 {
-    char buf[1024];
-
-    if (terminaltype == NULL)
-	return(1);
-
-    /*
-     * tgetent() will return 1 if the type is known, and
-     * 0 if it is not known.  If it returns -1, it couldn't
-     * open the database.  But if we can't open the database,
-     * it won't help to say we failed, because we won't be
-     * able to verify anything else.  So, we treat -1 like 1.
-     */
-    if (tgetent(buf, s) == 0)
-	return(0);
-    return(1);
+    return 1;
 }
 
 
@@ -672,14 +659,21 @@ char remote_host_name[MaxHostNameLen];
 /*
  * Get a pty, scan input lines.
  */
-void
-doit(struct sockaddr_in *who)
+static void
+doit(struct sockaddr *who, int who_len)
 {
-    char *host;
+    char *host = NULL;
     struct hostent *hp;
     int level;
     int ptynum;
     char user_name[256];
+    int error;
+    char host_addr[256];
+    void *addr;
+    int addr_sz;
+    const char *tmp;
+    int af;
+    struct sockaddr_in mapped_sin;
 
     /*
      * Find an available pty to use.
@@ -704,35 +698,57 @@ doit(struct sockaddr_in *who)
     }
 #endif	/* _SC_CRAY_SECURE_SYS */
 
-    /* get name of connected client */
-    hp = roken_gethostbyaddr((const char *)&who->sin_addr,
-		       sizeof (struct in_addr),
-		       who->sin_family);
+    af = who->sa_family;
+    switch (af) {
+    case AF_INET : {
+	struct sockaddr_in *sin = (struct sockaddr_in *)who;
+
+	addr    = &sin->sin_addr;
+	addr_sz = sizeof(sin->sin_addr);
+	break;
+    }
+#ifdef HAVE_IPV6
+    case AF_INET6 : {
+	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)who;
+
+	addr    = &sin6->sin6_addr;
+	addr_sz = sizeof(sin6->sin6_addr);
+	break;
+    }
+#endif
+    default :
+	fatal (net, "Unknown address family\r\n");
+	break;
+    }
+
+    hp = getipnodebyaddr (addr, addr_sz, af, &error);
 
     if (hp == NULL && registerd_host_only) {
 	fatal(net, "Couldn't resolve your address into a host name.\r\n\
 Please contact your net administrator");
-    } else if (hp) {
+    } else if (hp != NULL) {
 	host = hp->h_name;
-    } else {
-	host = inet_ntoa(who->sin_addr);
     }
+
+    tmp = inet_ntop(af, addr, host_addr, sizeof(host_addr));
+    if (tmp == NULL)
+	strcpy_truncate (host_addr, "unknown address", sizeof(host_addr));
+
+    if (host == NULL)
+	host = host_addr;
+
     /*
      * We must make a copy because Kerberos is probably going
      * to also do a gethost* and overwrite the static data...
      */
-    strncpy(remote_host_name, host, sizeof(remote_host_name)-1);
-    remote_host_name[sizeof(remote_host_name)-1] = 0;
+    strcpy_truncate(remote_host_name, host, sizeof(remote_host_name));
+    freehostent (hp);
     host = remote_host_name;
 
     /* XXX - should be k_gethostname? */
     gethostname(host_name, sizeof (host_name));
     hostname = host_name;
 
-#ifndef abs
-#define abs(x) ((x < 0) ? (-x) : x)
-#endif
-
     /* Only trim if too long (and possible) */
     if (strlen(remote_host_name) > abs(utmp_len)) {
 	char *domain = strchr(host_name, '.');
@@ -746,9 +762,9 @@ Please contact your net administrator");
      * If hostname still doesn't fit utmp, use ipaddr.
      */
     if (strlen(remote_host_name) > abs(utmp_len))
-	strncpy(remote_host_name,
-		inet_ntoa(who->sin_addr),
-		sizeof(remote_host_name)-1);
+	strcpy_truncate(remote_host_name,
+			host_addr,
+			sizeof(remote_host_name));
 
 #ifdef AUTHENTICATION
     auth_encrypt_init(hostname, host, "TELNETD", 1);
@@ -759,7 +775,7 @@ Please contact your net administrator");
      * get terminal type.
      */
     *user_name = 0;
-    level = getterminaltype(user_name);
+    level = getterminaltype(user_name, sizeof(user_name));
     setenv("TERM", terminaltype ? terminaltype : "network", 1);
 
 #ifdef _SC_CRAY_SECURE_SYS
@@ -771,26 +787,39 @@ Please contact your net administrator");
     }
 #endif	/* _SC_CRAY_SECURE_SYS */
 
-    telnet(net, ourpty, host, level, user_name);  /* begin server processing */
+    /* begin server processing */
+    my_telnet(net, ourpty, host, level, user_name);
     /*NOTREACHED*/
 }  /* end of doit */
 
+/* output contents of /etc/issue.net, or /etc/issue */
+static void
+show_issue(void)
+{
+    FILE *f;
+    char buf[128];
+    f = fopen("/etc/issue.net", "r");
+    if(f == NULL)
+	f = fopen("/etc/issue", "r");
+    if(f){
+	while(fgets(buf, sizeof(buf)-2, f)){
+	    strcpy(buf + strcspn(buf, "\r\n"), "\r\n");
+	    writenet((unsigned char*)buf, strlen(buf));
+	}
+	fclose(f);
+    }
+}
+
 /*
  * Main loop.  Select from pty and network, and
  * hand data to telnet receiver finite state machine.
  */
 void
-telnet(int f, int p, char *host, int level, char *autoname)
+my_telnet(int f, int p, char *host, int level, char *autoname)
 {
     int on = 1;
-#define	TABBUFSIZ	512
-    char	defent[TABBUFSIZ];
-    char	defstrs[TABBUFSIZ];
-#undef	TABBUFSIZ
     char *he;
-    char *HN;
     char *IM;
-    void netflush();
     int nfd;
     int startslave_called = 0;
     time_t timeout;
@@ -931,6 +960,7 @@ telnet(int f, int p, char *host, int level, char *autoname)
     }
 #endif
 
+    show_issue();
     /*
      * Show banner that getty never gave.
      *
