@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -38,12 +38,12 @@
 
 #include "krb5_locl.h"
 
-RCSID("$Id: get_in_tkt.c,v 1.67 1997/10/29 01:30:13 assar Exp $");
+RCSID("$Id: get_in_tkt.c,v 1.89 1999/07/04 15:42:35 assar Exp $");
 
 krb5_error_code
 krb5_init_etype (krb5_context context,
 		 unsigned *len,
-		 unsigned **val,
+		 int **val,
 		 const krb5_enctype *etypes)
 {
     int i;
@@ -63,8 +63,8 @@ krb5_init_etype (krb5_context context,
     for (i = 0; tmp[i]; ++i)
 	;
     *len = i;
-    *val = malloc(i * sizeof(unsigned));
-    if (*val == NULL) {
+    *val = malloc(i * sizeof(int));
+    if (i != 0 && *val == NULL) {
 	ret = ENOMEM;
 	goto cleanup;
     }
@@ -80,34 +80,42 @@ cleanup:
 
 static krb5_error_code
 decrypt_tkt (krb5_context context,
-	     const krb5_keyblock *key,
+	     krb5_keyblock *key,
+	     unsigned usage,
 	     krb5_const_pointer decrypt_arg,
 	     krb5_kdc_rep *dec_rep)
 {
     krb5_error_code ret;
     krb5_data data;
     size_t size;
+    krb5_crypto crypto;
+
+    krb5_crypto_init(context, key, 0, &crypto);
+
+    ret = krb5_decrypt_EncryptedData (context,
+				      crypto,
+				      usage,
+				      &dec_rep->kdc_rep.enc_part,
+				      &data);
+    krb5_crypto_destroy(context, crypto);
 
-    ret = krb5_decrypt (context,
-			dec_rep->part1.enc_part.cipher.data,
-			dec_rep->part1.enc_part.cipher.length,
-			dec_rep->part1.enc_part.etype,
-			key,
-			&data);
     if (ret)
 	return ret;
 
-    ret = decode_EncASRepPart(data.data,
-			      data.length,
-			      &dec_rep->part2, 
-			      &size);
-    if (ret)
-	ret = decode_EncTGSRepPart(data.data,
+    ret = krb5_decode_EncASRepPart(context,
+				   data.data,
 				   data.length,
-				   &dec_rep->part2, 
+				   &dec_rep->enc_part, 
 				   &size);
+    if (ret)
+	ret = krb5_decode_EncTGSRepPart(context,
+					data.data,
+					data.length,
+					&dec_rep->enc_part, 
+					&size);
     krb5_data_free (&data);
-    if (ret) return ret;
+    if (ret)
+	return ret;
     return 0;
 }
 
@@ -117,13 +125,14 @@ _krb5_extract_ticket(krb5_context context,
 		     krb5_creds *creds,		
 		     krb5_keyblock *key,
 		     krb5_const_pointer keyseed,
+		     krb5_key_usage key_usage,
 		     krb5_addresses *addrs,
 		     unsigned nonce,
 		     krb5_boolean allow_server_mismatch,
 		     krb5_decrypt_proc decrypt_proc,
 		     krb5_const_pointer decryptarg)
 {
-    krb5_error_code err;
+    krb5_error_code ret;
     krb5_principal tmp_principal;
     int tmp;
     time_t tmp_time;
@@ -131,26 +140,30 @@ _krb5_extract_ticket(krb5_context context,
 
     /* compare client */
 
-    err = principalname2krb5_principal (&tmp_principal,
-					rep->part1.cname,
-					rep->part1.crealm);
-    if (err)
+    ret = principalname2krb5_principal (&tmp_principal,
+					rep->kdc_rep.cname,
+					rep->kdc_rep.crealm);
+    if (ret)
 	goto out;
     tmp = krb5_principal_compare (context, tmp_principal, creds->client);
     krb5_free_principal (context, tmp_principal);
     if (!tmp) {
-	err = KRB5KRB_AP_ERR_MODIFIED;
+	ret = KRB5KRB_AP_ERR_MODIFIED;
 	goto out;
     }
     
     /* extract ticket */
     {
-	unsigned char buf[1024];
+	unsigned char *buf;
 	size_t len;
-	encode_Ticket(buf + sizeof(buf) - 1, sizeof(buf), 
-			    &rep->part1.ticket, &len);
-	creds->ticket.data = malloc(len);
-	memcpy(creds->ticket.data, buf + sizeof(buf) - len, len);
+	len = length_Ticket(&rep->kdc_rep.ticket);
+	buf = malloc(len);
+	if(buf == NULL) {
+	    ret = ENOMEM;
+	    goto out;
+	}
+	encode_Ticket(buf + len - 1, len, &rep->kdc_rep.ticket, &len);
+	creds->ticket.data = buf;
 	creds->ticket.length = len;
 	creds->second_ticket.length = 0;
 	creds->second_ticket.data   = NULL;
@@ -158,10 +171,10 @@ _krb5_extract_ticket(krb5_context context,
 
     /* compare server */
 
-    err = principalname2krb5_principal (&tmp_principal,
-					rep->part1.ticket.sname,
-					rep->part1.ticket.realm);
-    if (err)
+    ret = principalname2krb5_principal (&tmp_principal,
+					rep->kdc_rep.ticket.sname,
+					rep->kdc_rep.ticket.realm);
+    if (ret)
 	goto out;
     if(allow_server_mismatch){
 	krb5_free_principal(context, creds->server);
@@ -171,7 +184,7 @@ _krb5_extract_ticket(krb5_context context,
 	tmp = krb5_principal_compare (context, tmp_principal, creds->server);
 	krb5_free_principal (context, tmp_principal);
 	if (!tmp) {
-	    err = KRB5KRB_AP_ERR_MODIFIED;
+	    ret = KRB5KRB_AP_ERR_MODIFIED;
 	    goto out;
 	}
     }
@@ -181,14 +194,21 @@ _krb5_extract_ticket(krb5_context context,
     if (decrypt_proc == NULL)
 	decrypt_proc = decrypt_tkt;
     
-    err = (*decrypt_proc)(context, key, decryptarg, rep);
-    if (err)
+    ret = (*decrypt_proc)(context, key, key_usage, decryptarg, rep);
+    if (ret)
 	goto out;
 
+#if 0
+    /* XXX should this decode be here, or in the decrypt_proc? */
+    ret = krb5_decode_keyblock(context, &rep->enc_part.key, 1);
+    if(ret)
+	goto out;
+#endif
+
     /* compare nonces */
 
-    if (nonce != rep->part2.nonce) {
-	err = KRB5KRB_AP_ERR_MODIFIED;
+    if (nonce != rep->enc_part.nonce) {
+	ret = KRB5KRB_AP_ERR_MODIFIED;
 	goto out;
     }
 
@@ -196,81 +216,81 @@ _krb5_extract_ticket(krb5_context context,
 
     krb5_timeofday (context, &sec_now);
     if (context->kdc_sec_offset == 0
-	&& krb5_config_get_bool (context->cf,
+	&& krb5_config_get_bool (context, NULL,
 				 "libdefaults",
 				 "kdc_timesync",
 				 NULL)) {
-	context->kdc_sec_offset = rep->part2.authtime - sec_now;
+	context->kdc_sec_offset = rep->enc_part.authtime - sec_now;
 	krb5_timeofday (context, &sec_now);
     }
 
     /* check all times */
 
-    if (rep->part2.starttime) {
-	tmp_time = *rep->part2.starttime;
+    if (rep->enc_part.starttime) {
+	tmp_time = *rep->enc_part.starttime;
     } else
-	tmp_time = rep->part2.authtime;
+	tmp_time = rep->enc_part.authtime;
 
     if (creds->times.starttime == 0
 	&& abs(tmp_time - sec_now) > context->max_skew) {
-	err = KRB5KRB_AP_ERR_MODIFIED;
+	ret = KRB5KRB_AP_ERR_SKEW;
 	goto out;
     }
 
     if (creds->times.starttime != 0
 	&& tmp_time != creds->times.starttime) {
-	err = KRB5KRB_AP_ERR_MODIFIED;
+	ret = KRB5KRB_AP_ERR_MODIFIED;
 	goto out;
     }
 
     creds->times.starttime = tmp_time;
 
-    if (rep->part2.renew_till) {
-	tmp_time = *rep->part2.renew_till;
+    if (rep->enc_part.renew_till) {
+	tmp_time = *rep->enc_part.renew_till;
     } else
 	tmp_time = 0;
 
     if (creds->times.renew_till != 0
 	&& tmp_time > creds->times.renew_till) {
-	err = KRB5KRB_AP_ERR_MODIFIED;
+	ret = KRB5KRB_AP_ERR_MODIFIED;
 	goto out;
     }
 
     creds->times.renew_till = tmp_time;
 
-    creds->times.authtime = rep->part2.authtime;
+    creds->times.authtime = rep->enc_part.authtime;
 
     if (creds->times.endtime != 0
-	&& rep->part2.endtime > creds->times.endtime) {
-	err = KRB5KRB_AP_ERR_MODIFIED;
+	&& rep->enc_part.endtime > creds->times.endtime) {
+	ret = KRB5KRB_AP_ERR_MODIFIED;
 	goto out;
     }
 
-    creds->times.endtime  = rep->part2.endtime;
+    creds->times.endtime  = rep->enc_part.endtime;
 
-    if(rep->part2.caddr)
-	krb5_copy_addresses (context, rep->part2.caddr, &creds->addresses);
+    if(rep->enc_part.caddr)
+	krb5_copy_addresses (context, rep->enc_part.caddr, &creds->addresses);
     else if(addrs)
 	krb5_copy_addresses (context, addrs, &creds->addresses);
     else {
 	creds->addresses.len = 0;
 	creds->addresses.val = NULL;
     }
-    creds->flags.b = rep->part2.flags;
+    creds->flags.b = rep->enc_part.flags;
 	  
-    creds->authdata.length = 0;
-    creds->authdata.data = NULL;
+    creds->authdata.len = 0;
+    creds->authdata.val = NULL;
     creds->session.keyvalue.length = 0;
     creds->session.keyvalue.data   = NULL;
-    creds->session.keytype = rep->part2.key.keytype;
-    err = krb5_data_copy (&creds->session.keyvalue,
-			  rep->part2.key.keyvalue.data,
-			  rep->part2.key.keyvalue.length);
+    creds->session.keytype = rep->enc_part.key.keytype;
+    ret = krb5_data_copy (&creds->session.keyvalue,
+			  rep->enc_part.key.keyvalue.data,
+			  rep->enc_part.key.keyvalue.length);
 
 out:
-    memset (rep->part2.key.keyvalue.data, 0,
-	    rep->part2.key.keyvalue.length);
-    return err;
+    memset (rep->enc_part.key.keyvalue.data, 0,
+	    rep->enc_part.key.keyvalue.length);
+    return ret;
 }
 
 
@@ -285,6 +305,7 @@ make_pa_enc_timestamp(krb5_context context, PA_DATA *pa,
     krb5_error_code ret;
     int32_t sec, usec;
     unsigned usec2;
+    krb5_crypto crypto;
     
     krb5_us_timeofday (context, &sec, &usec);
     p.patimestamp = sec;
@@ -298,19 +319,15 @@ make_pa_enc_timestamp(krb5_context context, PA_DATA *pa,
     if (ret)
 	return ret;
 
-    /*
-     * According to the spec this is the only encryption method
-     * that must be supported so it's the safest choice.  On the
-     * other hand, old KDCs might not support it.
-     */
-
+    krb5_crypto_init(context, key, 0, &crypto);
     ret = krb5_encrypt_EncryptedData(context, 
+				     crypto,
+				     KRB5_KU_PA_ENC_TIMESTAMP,
 				     buf + sizeof(buf) - len,
 				     len,
-				     etype,
 				     0,
-				     key,
 				     &encdata);
+    krb5_crypto_destroy(context, crypto);
     if (ret)
 	return ret;
 		    
@@ -329,6 +346,56 @@ make_pa_enc_timestamp(krb5_context context, PA_DATA *pa,
     return 0;
 }
 
+static krb5_error_code
+add_padata(krb5_context context,
+	   METHOD_DATA *md, 
+	   krb5_principal client,
+	   krb5_key_proc key_proc,
+	   krb5_const_pointer keyseed,
+	   krb5_enctype *enctypes, 
+	   unsigned netypes,
+	   krb5_salt *salt)
+{
+    krb5_error_code ret;
+    PA_DATA *pa2;
+    krb5_salt salt2;
+    krb5_enctype *ep;
+    int i;
+    
+    if(salt == NULL) {
+	/* default to standard salt */
+	ret = krb5_get_pw_salt (context, client, &salt2);
+	salt = &salt2;
+    }
+    if (!enctypes) {
+	enctypes = context->etypes; /* XXX */
+	netypes = 0;
+	for (ep = enctypes; *ep != ETYPE_NULL; ep++)
+	    netypes++;
+    }
+    pa2 = realloc (md->val, (md->len + netypes) * sizeof(*md->val));
+    if (pa2 == NULL)
+	return ENOMEM;
+    md->val = pa2;
+
+    for (i = 0; i < netypes; ++i) {
+	krb5_keyblock *key;
+
+	ret = (*key_proc)(context, enctypes[i], *salt, keyseed, &key);
+	if (ret)
+	    continue;
+	ret = make_pa_enc_timestamp (context, &md->val[md->len],
+				     enctypes[i], key);
+	krb5_free_keyblock (context, key);
+	if (ret)
+	    return ret;
+	++md->len;
+    }
+    if(salt == &salt2)
+	krb5_free_salt(context, salt2);
+    return 0;
+}
+
 static krb5_error_code
 init_as_req (krb5_context context,
 	     krb5_kdc_flags opts,
@@ -336,15 +403,14 @@ init_as_req (krb5_context context,
 	     const krb5_addresses *addrs,
 	     const krb5_enctype *etypes,
 	     const krb5_preauthtype *ptypes,
+	     const krb5_preauthdata *preauth,
 	     krb5_key_proc key_proc,
 	     krb5_const_pointer keyseed,
 	     unsigned nonce,
 	     AS_REQ *a)
 {
     krb5_error_code ret;
-    krb5_data salt;
-    krb5_keyblock *key;
-    krb5_enctype etype;
+    krb5_salt salt;
 
     memset(a, 0, sizeof(*a));
 
@@ -399,83 +465,93 @@ init_as_req (krb5_context context,
     if (ret)
 	goto fail;
 
-    etype = a->req_body.etype.val[0]; /* XXX */
+    /*
+     * This means no addresses
+     */
 
-    a->req_body.addresses = malloc(sizeof(*a->req_body.addresses));
-    if (a->req_body.addresses == NULL) {
-	ret = ENOMEM;
-	goto fail;
+    if (addrs && addrs->len == 0) {
+	a->req_body.addresses = NULL;
+    } else {
+	a->req_body.addresses = malloc(sizeof(*a->req_body.addresses));
+	if (a->req_body.addresses == NULL) {
+	    ret = ENOMEM;
+	    goto fail;
+	}
+
+	if (addrs)
+	    ret = krb5_copy_addresses(context, addrs, a->req_body.addresses);
+	else
+	    ret = krb5_get_all_client_addrs (context, a->req_body.addresses);
+	if (ret)
+	    return ret;
     }
 
-    if (addrs)
-	ret = krb5_copy_addresses(context, addrs, a->req_body.addresses);
-    else
-	ret = krb5_get_all_client_addrs (a->req_body.addresses);
-    if (ret)
-	return ret;
-
     a->req_body.enc_authorization_data = NULL;
     a->req_body.additional_tickets = NULL;
 
+    if(preauth != NULL) {
+	int i;
+	ALLOC(a->padata, 1);
+	if(a->padata == NULL) {
+	    ret = ENOMEM;
+	    goto fail;
+	}
+	for(i = 0; i < preauth->len; i++) {
+	    if(preauth->val[i].type == KRB5_PADATA_ENC_TIMESTAMP){
+		int j;
+		PA_DATA *tmp = realloc(a->padata->val, 
+				       (a->padata->len + 
+					preauth->val[i].info.len) * 
+				       sizeof(*a->padata->val));
+		if(tmp == NULL) {
+		    ret = ENOMEM;
+		    goto fail;
+		}
+		a->padata->val = tmp;
+		for(j = 0; j < preauth->val[i].info.len; j++) {
+		    krb5_salt *sp = &salt;
+		    if(preauth->val[i].info.val[j].salttype)
+			salt.salttype = *preauth->val[i].info.val[j].salttype;
+		    else
+			salt.salttype = KRB5_PW_SALT;
+		    if(preauth->val[i].info.val[j].salt)
+			salt.saltvalue = *preauth->val[i].info.val[j].salt;
+		    else
+			if(salt.salttype == KRB5_PW_SALT)
+			    sp = NULL;
+			else
+			    krb5_data_zero(&salt.saltvalue);
+		    add_padata(context, a->padata, creds->client, 
+			       key_proc, keyseed, 
+			       &preauth->val[i].info.val[j].etype, 1,
+			       sp);
+		}
+	    }
+	}
+    } else 
     /* not sure this is the way to use `ptypes' */
     if (ptypes == NULL || *ptypes == KRB5_PADATA_NONE)
 	a->padata = NULL;
     else if (*ptypes ==  KRB5_PADATA_ENC_TIMESTAMP) {
-	a->padata = malloc(sizeof(*a->padata));
+	ALLOC(a->padata, 1);
 	if (a->padata == NULL) {
 	    ret = ENOMEM;
 	    goto fail;
 	}
-	    
-	a->padata->len = 2;
-	a->padata->val = calloc(a->padata->len, sizeof(*a->padata->val));
-	if (a->padata->val == NULL) {
-	    ret = ENOMEM;
-	    goto fail;
-	}
-	
+	a->padata->len = 0;
+	a->padata->val = NULL;
+
 	/* make a v5 salted pa-data */
-	salt.length = 0;
-	salt.data = NULL;
-	ret = krb5_get_salt (creds->client, &salt);
+	add_padata(context, a->padata, creds->client, 
+		   key_proc, keyseed, a->req_body.etype.val,
+		   a->req_body.etype.len, NULL);
 	
-	if (ret)
-	    goto fail;
-	
-	{
-	    krb5_keytype keytype;
-	    ret = krb5_etype_to_keytype(context, etype, &keytype);
-	    if(ret){
-		krb5_data_free(&salt);
-		goto fail;
-	    }
-	    ret = (*key_proc)(context, keytype, &salt, keyseed, &key);
-	    krb5_data_free (&salt);
-	}
-	if (ret)
-	    goto fail;
-	ret = make_pa_enc_timestamp(context, &a->padata->val[0], etype, key);
-	krb5_free_keyblock (context, key);
-	free (key);
-	if (ret)
-	    goto fail;
 	/* make a v4 salted pa-data */
-	salt.length = 0;
-	salt.data = NULL;
-	{
-	    krb5_keytype keytype;
-	    ret = krb5_etype_to_keytype(context, etype, &keytype);
-	    if(ret)
-		goto fail;
-	    ret = (*key_proc)(context, keytype, &salt, keyseed, &key);
-	}
-	if (ret)
-	    goto fail;
-	ret = make_pa_enc_timestamp(context, &a->padata->val[1], etype, key);
-	krb5_free_keyblock (context, key);
-	free (key);
-	if (ret)
-	    goto fail;
+	salt.salttype = KRB5_PW_SALT;
+	krb5_data_zero(&salt.saltvalue);
+	add_padata(context, a->padata, creds->client, 
+		   key_proc, keyseed, a->req_body.etype.val,
+		   a->req_body.etype.len, &salt);
     } else {
 	ret = KRB5_PREAUTH_BAD_TYPE;
 	goto fail;
@@ -492,6 +568,7 @@ krb5_get_in_cred(krb5_context context,
 		 const krb5_addresses *addrs,
 		 const krb5_enctype *etypes,
 		 const krb5_preauthtype *ptypes,
+		 const krb5_preauthdata *preauth,
 		 krb5_key_proc key_proc,
 		 krb5_const_pointer keyseed,
 		 krb5_decrypt_proc decrypt_proc,
@@ -504,7 +581,7 @@ krb5_get_in_cred(krb5_context context,
     krb5_kdc_rep rep;
     krb5_data req, resp;
     char buf[BUFSIZ];
-    krb5_data salt;
+    krb5_salt salt;
     krb5_keyblock *key;
     size_t size;
     krb5_kdc_flags opts;
@@ -523,6 +600,7 @@ krb5_get_in_cred(krb5_context context,
 		       addrs,
 		       etypes,
 		       ptypes,
+		       preauth,
 		       key_proc,
 		       keyseed,
 		       nonce,
@@ -545,7 +623,7 @@ krb5_get_in_cred(krb5_context context,
 	return ret;
 
     memset (&rep, 0, sizeof(rep));
-    if((ret = decode_AS_REP(resp.data, resp.length, &rep.part1, &size))){
+    if((ret = decode_AS_REP(resp.data, resp.length, &rep.kdc_rep, &size))) {
 	/* let's try to parse it as a KRB-ERROR */
 	KRB_ERROR error;
 	int ret2;
@@ -555,47 +633,63 @@ krb5_get_in_cred(krb5_context context,
 	    ret = KRB5KRB_AP_ERR_V4_REPLY;
 	krb5_data_free(&resp);
 	if (ret2 == 0) {
-	    /* XXX - send krb-error packet to caller */
-	    free_KRB_ERROR (&error);
-	    return error.error_code;
+	    ret = error.error_code;
+	    if(ret_as_reply)
+		ret_as_reply->error = error;
+	    else
+		free_KRB_ERROR (&error);
+	    return ret;
 	}
 	return ret;
     }
     krb5_data_free(&resp);
     
     pa = NULL;
-    etype = rep.part1.enc_part.etype;
-    if(rep.part1.padata){
+    etype = rep.kdc_rep.enc_part.etype;
+    if(rep.kdc_rep.padata){
 	int index = 0;
-	pa = krb5_find_padata(rep.part1.padata->val, rep.part1.padata->len, 
+	pa = krb5_find_padata(rep.kdc_rep.padata->val, rep.kdc_rep.padata->len, 
 			      pa_pw_salt, &index);
+	if(pa == NULL) {
+	    index = 0;
+	    pa = krb5_find_padata(rep.kdc_rep.padata->val, 
+				  rep.kdc_rep.padata->len, 
+				  pa_afs3_salt, &index);
+	}
     }
     if(pa) {
-	krb5_keytype keytype;
-	ret = krb5_etype_to_keytype(context, etype, &keytype);
-	ret = (*key_proc)(context, keytype, &pa->padata_value, keyseed, &key);
+	salt.salttype = pa->padata_type;
+	salt.saltvalue = pa->padata_value;
+	
+	ret = (*key_proc)(context, etype, salt, keyseed, &key);
     } else {
 	/* make a v5 salted pa-data */
-	krb5_keytype keytype;
-	salt.length = 0;
-	salt.data = NULL;
-	ret = krb5_get_salt (creds->client, &salt);
+	ret = krb5_get_pw_salt (context, creds->client, &salt);
 	
 	if (ret)
-	    return ret;
-	ret = krb5_etype_to_keytype(context, etype, &keytype);
-	ret = (*key_proc)(context, keytype, &salt, keyseed, &key);
-	krb5_data_free (&salt);
+	    goto out;
+	ret = (*key_proc)(context, etype, salt, keyseed, &key);
+	krb5_free_salt(context, salt);
     }
     if (ret)
-	return ret;
+	goto out;
 	
-    ret = _krb5_extract_ticket(context, &rep, creds, key, keyseed, 
-			      NULL, nonce, FALSE, decrypt_proc, decryptarg);
+    ret = _krb5_extract_ticket(context, 
+			       &rep, 
+			       creds, 
+			       key, 
+			       keyseed, 
+			       KRB5_KU_AS_REP_ENC_PART,
+			       NULL, 
+			       nonce, 
+			       FALSE, 
+			       decrypt_proc, 
+			       decryptarg);
     memset (key->keyvalue.data, 0, key->keyvalue.length);
-    krb5_free_keyblock (context, key);
+    krb5_free_keyblock_contents (context, key);
     free (key);
 
+out:
     if (ret == 0 && ret_as_reply)
 	*ret_as_reply = rep;
     else
@@ -618,12 +712,16 @@ krb5_get_in_tkt(krb5_context context,
 		krb5_kdc_rep *ret_as_reply)
 {
     krb5_error_code ret;
-
+    krb5_kdc_flags opts;
+    opts.i = 0;
+    opts.b = int2KDCOptions(options);
+    
     ret = krb5_get_in_cred (context,
-			    options,
+			    opts.i,
 			    addrs,
 			    etypes,
 			    ptypes,
+			    NULL,
 			    key_proc,
 			    keyseed,
 			    decrypt_proc,
